from __future__ import annotations
import glob
import copy
import numpy as np
import numpy.linalg as LA
import matplotlib.pyplot as plt


class TdObj:

    def __init__(self, name: str = "No Name", meta: dict = None,
                 data: dict = None, max_value: float = None, datalen: int = 1000):

        self.meta = meta
        self.data = data
        self.max_value = max_value
        self.datalen = datalen
        self.name = name
        self.trials_count = len(meta["DURATION"])

        if meta is None or data is None or max_value is None or self.trials_count != len(meta["DELAY"]):
            raise Exception("Data may be corrupted")

    def __add__(self, other) -> TdObj:

        new_meta: dict = copy.deepcopy(self.meta)
        new_data: dict = copy.deepcopy(self.data)
        new_max_value: float = self.max_value

        if isinstance(other, TdObj):
            if len(self.meta) != len(other.meta) or len(self.data) != len(other.data) or self.datalen != other.datalen:
                print("Different formats are mixed")

            metaKeys = self.meta.keys()
            for key in metaKeys:
                new_meta[key] += other.meta[key]

            dataKeys = self.data.keys()
            for key in dataKeys:
                new_data[key] += other.data[key]

            new_max_value += other.max_value

        return TdObj(meta=new_meta, data=new_data, max_value=new_max_value, datalen=self.datalen)

    def __radd__(self, other) -> TdObj:
        return self + other

    def __iadd__(self, other) -> TdObj:
        self = self + other
        return self

    def setName(self, name: str) -> TdObj:
        self.name = name
        return self

    def getAverageDuration(self) -> float:
        return sum(self.meta["DURATION"]) / self.trials_count

    def getAverageDelay(self) -> float:
        return sum(self.meta["NORMALIZED_DELAY"]) / len(self.meta["NORMALIZED_DELAY"])

    def normalizeValue(self, max_value: float = 1.0) -> TdObj:
        dataKeys = self.data.keys()
        for key in dataKeys:
            self.data[key] *= max_value / self.max_value

        self.max_value = max_value
        return self

    def convolve(self, weight: int) -> TdObj:
        dataKeys = self.data.keys()
        for key in dataKeys:
            edge = int(weight / 2)
            edged = np.concatenate([[0] * edge, self.data[key], [self.data[key][-1]] * (weight - edge - 1)])
            self.data[key] = np.convolve(edged, [1 / weight] * weight, "valid")

        return self

    def smooth(self, weight: int = 1) -> TdObj:
        for _ in range(weight):
            self.convolve(3)

        return self

    def draw(self, layout: list = None, scale: float = 1.0, size: float = (12, 8), font: str = "DejaVu Sans", show: bool = True,
             auto_adjust: bool = True, show_legend: bool = True, show_chance: bool = True, show_signif: bool = True,
             show_total: bool = False, show_average_delay: bool = False) -> TdObj:

        if layout is None:
            layout = [[list(self.data.keys())]]

        rows = len(layout)
        columns = max(len(row) for row in layout)
        axes = [[None] * columns] * rows
        count = 1
        standard_ax = np.arange(0, self.datalen) / self.datalen * scale

        fig = plt.figure(figsize=size)
        fig.suptitle(self.name)

        for i, row in enumerate(layout):
            for j, column in enumerate(row):
                axes[i][j] = (fig.add_subplot(rows, columns, count, xlabel="Normalized Time", ylabel="Dominance Proportion")
                              if auto_adjust else
                              fig.add_subplot(rows, columns, count, xlabel="Normalized Time", ylabel="Dominance Proportion",
                                              ylim=(0, self.max_value)))
                count += 1
                for key in column:
                    axes[i][j].plot(standard_ax, self.data[key], label=key)

                # graph parameters
                chance = 1 / len(self.data)
                signif = chance + 1.645 * np.sqrt(chance * (1 - chance) / self.trials_count)
                average_delay = self.getAverageDelay() / self.datalen * scale

                show_chance and axes[i][j].plot(
                    [0, scale], [chance * self.max_value] * 2, label="Chance", linestyle="dotted", color="gray")
                show_signif and axes[i][j].plot(
                    [0, scale], [signif * self.max_value] * 2, label="Signif.", linestyle="dashed", color="gray")
                show_total and axes[i][j].plot(
                    standard_ax, sum(self.data[key] for key in self.data), label="Total", linestyle="dashed")

                show_average_delay and axes[i][j].plot([average_delay, average_delay + scale / self.datalen],
                                                       [0, self.max_value],
                                                       label="Average Delay",
                                                       linestyle="dotted")

                show_legend and axes[i][j].legend(prop={"family": font})

        # show graph or not
        show and plt.show()

        return self


class PcaObj:

    def __init__(self, label: list, data: np.ndarray, number_of_pcs: int = 2):
        self.label = label
        self.data = data
        self.corrcoef = np.corrcoef(self.data)

        # eigenvalues, eigenvectors
        w, v = LA.eig(self.corrcoef)
        v = v.T
        index = np.argsort(w)[::-1]

        self.eigenvalues = np.array([w[i] for i in index])
        self.eigenvectors = np.array([v[i] for i in index])

        print("Eigenvalues:\n", self.eigenvalues, "\n")
        print("Eigenvectors:\n", self.eigenvectors, "\n")

        # pc scores and vectors for each item
        self.pc = np.array([self.data.T @ self.eigenvectors[i] for i in range(number_of_pcs)])
        self.vpc = np.array([np.identity(len(self.eigenvectors.T)) @ self.eigenvectors[i] for i in range(number_of_pcs)])

        for i in range(number_of_pcs):
            print("Score of PC" + str(i + 1) + ":\n", self.pc[i], "\n")

    def draw(self, size: float = (6, 4), font: str = "DejaVu Sans",
             show: bool = True, show_legend: bool = True) -> PcaObj:

        for i in range(len(self.pc) - 1):
            for j in range(i + 1, len(self.pc)):
                xlabel = "PC" + str(i + 1)
                ylabel = "PC" + str(j + 1)

                fig = plt.figure(figsize=size)
                fig.suptitle(xlabel + " and " + ylabel)
                ax = fig.add_subplot(xlabel=xlabel, ylabel=ylabel)
                ax.scatter(self.pc[i], self.pc[j])

                for k in range(len(self.label)):
                    ax.plot([0, self.vpc[i][k]], [0, self.vpc[j][k]], label=self.label[k])

                ax.grid()
                show_legend and ax.legend(prop={"family": font})

        show and plt.show()
        return self


class TdContainer:

    def __init__(self, contents: list[TdObj]):
        self.contents = contents
        self.likelihoods = None

    def __getitem__(self, key) -> TdObj:
        return self.contents[key]

    def merge(self) -> TdObj:
        result: TdObj = None
        for obj in self.contents:
            result += obj

        return result.normalizeValue(1)

    def mergeAs(self, name) -> TdObj:
        return self.merge().setName(name)

    def pca(self, number_of_pcs: int = 2) -> PcaObj:
        label: list = list(self.contents[0].meta["DOMINANCE_DURATION"][0].keys())
        table: list = []
        for individual in self.contents:
            for elem in individual.meta["DOMINANCE_DURATION"]:
                data = [None] * len(label)
                for key in elem:
                    data[label.index(key)] = elem[key]

                table.append(data)

        data_matrix: np.ndarray = np.array(table).T

        # normalize
        for i in range(len(data_matrix)):
            data: np.ndarray = data_matrix[i]
            data_matrix[i] = (data - data.mean()) / np.abs(data.std())

        return PcaObj(label=label, data=data_matrix, number_of_pcs=number_of_pcs)

    def likelihood(self) -> list[float]:
        if self.likelihoods is not None:
            return self.likelihoods

        result: list[float] = [1] * len(self.contents)
        merged_data: dict = self.merge().normalizeValue(1).data
        for i in range(len(self.contents)):
            current_obj: TdObj = self.contents[i]
            for key in merged_data.keys():
                p1: np.ndarray = merged_data[key]
                p2: np.ndarray = current_obj.data[key]
                result[i] *= ((p1 @ p2) + (1 - p1) @ (1 - p2)) / (current_obj.datalen * current_obj.max_value)

        self.likelihoods = copy.deepcopy(result)

        return result

    def boxPlot(self, size: float = (6, 4), show: bool = True) -> TdContainer:
        fig = plt.figure(figsize=size)
        fig.suptitle("Box Plot")
        ax = fig.add_subplot(ylabel="Likelihood")
        ax.set_xticklabels(["this"])
        ax.boxplot(self.likelihood())

        ax.scatter([0.99] * len(self.contents), self.likelihood(), s=10)
        show and plt.show()

        return self


def getDataFromStr(tab_separated_values: str, max_value: float = 1.0, datalen: int = 1000) -> TdObj:
    data: dict = {}
    meta: dict = {}
    time: dict = {}
    lines: list = tab_separated_values.replace("\r\n", "\n").replace("\r", "\n").split("\n")

    for line in lines:
        params: list = line.split("\t")
        if len(params) > 1:
            if line[0] == "#":
                tag = params[0][1:].upper()
                meta[tag] = [float(param) for param in params[1:]] if tag == "DURATION" or tag == "DELAY" else params[1:]
            else:
                time[params[0]] = [float(param) for param in params[1:]]

        elif len(params) == 1 and params[0] != "":
            if line[0] == "#":
                meta[params[0][1:].upper()] = []
            else:
                time[params[0]] = []

    if not ("DURATION" in meta):
        meta["DURATION"] = [max(max(time[key]) for key in time)]
        print("The duration data was not found, so it was automatically generated")

    if not ("DELAY" in meta):
        meta["DELAY"] = [min(min(time[key]) for key in time)]
        print("The delay data was not found, so it was automatically generated")

    duration: float = meta["DURATION"][0]
    delay: float = meta["DELAY"][0]
    dominance_durs: dict = {}

    for key in time:
        sample: list = time[key]
        data[key] = np.zeros(datalen, float)
        dominance_duration: float = 0.0
        for i in range(0, len(sample), 2):
            timebegin = sample[i]
            timeend = sample[i + 1]
            dominance_duration += timeend - timebegin
            for j in range(int(timebegin / duration * datalen), int(timeend / duration * datalen)):
                data[key][j] = max_value

        dominance_durs[key] = dominance_duration / duration

    meta["NORMALIZED_DELAY"] = [delay / duration * datalen]
    meta["DOMINANCE_DURATION"] = [dominance_durs]

    return TdObj(meta=meta, data=data, max_value=max_value, datalen=datalen)


def loadFile(file_path: str, max_value: float = 1.0, datalen: int = 1000) -> TdObj:
    with open(file_path, "r", encoding="UTF-8") as tdDataFile:
        try:
            return getDataFromStr(tdDataFile.read(), max_value=max_value, datalen=datalen)
        except Exception as e:
            print(e)

    return None


def loadFolder(folder_path: str = ".", file_extension: str = ".csv",
               max_value: float = 1.0, datalen: int = 1000) -> TdContainer:

    TdObj_list: list = []
    files: list = glob.glob(folder_path + "/*" + file_extension)

    for f in files:
        gotObj = loadFile(f, max_value=max_value, datalen=datalen)
        if gotObj is not None:
            TdObj_list.append(gotObj)

    if len(TdObj_list) == 0:
        print("Folder seems to be empty")

    return TdContainer(TdObj_list)
